import datetime
import glob
import os
import re
import shutil
import geopandas
import warnings
from importlib.resources import files

import backoff
import geopandas as gpd
import h5py
import httpx
import numpy as np
import pandas as pd
import rasterio
from httpx import HTTPError
from rasterio.mask import mask
from rasterio.merge import merge
from rasterio.transform import from_origin
from rasterstats import zonal_stats
from tqdm.auto import tqdm

from . import logger


def file_to_raster(f, variable, output_path, quality_flag_rm):
    h5_data = h5py.File(f, "r")

    # Daily --------------------------------------------------
    if ("VNP46A1" in f) or ("VNP46A2" in f):
        #### Check
        h5_names = list(
            h5_data["HDFEOS"]["GRIDS"]["VNP_Grid_DNB"]["Data Fields"].keys()
        )

        if variable not in h5_names:
            warnings.warn(
                "'"
                + variable
                + "'"
                + " not a valid variable option. Valid options include:\n"
                + ",\n ".join(h5_names),
                UserWarning,
            )

        tile_i = re.findall(r"h\d{2}v\d{2}", f)[0]

        bm_tiles_sf = gpd.read_file(
            files("blackmarble.data").joinpath("blackmarbletiles.geojson")
        )
        grid_i_sf = bm_tiles_sf[bm_tiles_sf["TileID"] == tile_i]

        xMin = float(grid_i_sf.geometry.bounds.minx.iloc[0])
        yMin = float(grid_i_sf.geometry.bounds.miny.iloc[0])
        xMax = float(grid_i_sf.geometry.bounds.maxx.iloc[0])
        yMax = float(grid_i_sf.geometry.bounds.maxy.iloc[0])

        out = h5_data["HDFEOS"]["GRIDS"]["VNP_Grid_DNB"]["Data Fields"][variable]
        qf = h5_data["HDFEOS"]["GRIDS"]["VNP_Grid_DNB"]["Data Fields"][
            "Mandatory_Quality_Flag"
        ]

        out = out[:]
        qf = qf[:]

        if len(quality_flag_rm) > 0:
            for val in quality_flag_rm:
                out = np.where(qf == val, np.nan, out)

    # Monthly / Annual --------------------------------------------------
    else:
        h5_names = list(
            h5_data["HDFEOS"]["GRIDS"]["VIIRS_Grid_DNB_2d"]["Data Fields"].keys()
        )

        if variable not in h5_names:
            warnings.warn(
                "'"
                + variable
                + "'"
                + " not a valid variable option. Valid options include:\n"
                + ",\n ".join(h5_names),
                UserWarning,
            )

        lat = h5_data["HDFEOS"]["GRIDS"]["VIIRS_Grid_DNB_2d"]["Data Fields"]["lat"]
        lon = h5_data["HDFEOS"]["GRIDS"]["VIIRS_Grid_DNB_2d"]["Data Fields"]["lon"]
        out = h5_data["HDFEOS"]["GRIDS"]["VIIRS_Grid_DNB_2d"]["Data Fields"][variable]

        out = out[:]

        #### Quality Flags
        if len(quality_flag_rm) > 0:
            variable_short = variable
            variable_short = re.sub("_Num", "", variable_short)
            variable_short = re.sub("_Std", "", variable_short)

            qf_name = variable_short + "_Quality"

            if qf_name in h5_names:
                qf = h5_data["HDFEOS"]["GRIDS"]["VIIRS_Grid_DNB_2d"]["Data Fields"][
                    variable + "_Quality"
                ]
                qf = qf[:]

                for val in quality_flag_rm:
                    out = np.where(qf == val, np.nan, out)

        xMin = min(lon)
        yMin = min(lat)
        xMax = max(lon)
        yMax = max(lat)

    # Cleanup --------------------------------------------------

    # Metadata
    nRows = out.shape[0]
    nCols = out.shape[1]
    res = nRows
    nodata_val = 65535
    myCrs = 4326

    # Makes raster
    # data = out[:]
    data = out
    # data = np.where(data == 65535, np.nan, data)

    # Define the pixel size and number of rows and columns
    pixel_size = 1  # Size of each pixel in the output raster
    rows, cols = data.shape

    # Define the spatial extent (bounding box) of the raster
    left, bottom, right, top = xMin, yMin, xMax, yMax

    psize_x = (xMax - xMin) / cols
    psize_y = (yMax - yMin) / rows

    # Create the raster file
    with rasterio.open(
        output_path,
        "w",
        driver="GTiff",
        height=rows,
        width=cols,
        count=1,
        dtype=data.dtype,
        crs="EPSG:4326",
        transform=from_origin(left, top, psize_x, psize_y),
    ) as dst:
        dst.write(data, 1)

    return None


@backoff.on_exception(
    backoff.expo,
    HTTPError,
)
def retrieve_manifest(product_id: str, date: datetime.date) -> pd.DataFrame:
    """NASA Black Marble

    Returns
    -------
    pandas.DataFrame
        NASA Black Marble data manifest (i.e., downloads links)
    """

    match product_id:
        case "VNP46A3":
            # if VNP46A3 then first day of the month
            tm_yday = date.replace(day=1).timetuple().tm_yday
        case "VNP46A4":
            # if VNP46A4 then first day of the year
            tm_yday = date.replace(month=1, day=1).timetuple().tm_yday
        case _:
            tm_yday = date.timetuple().tm_yday

    manifest = pd.read_csv(
        f"https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/5000/{product_id}/{date.year}/{tm_yday}.csv"
    )
    manifest["year"] = date.year
    manifest["day"] = tm_yday

    return manifest


@backoff.on_exception(
    backoff.expo,
    HTTPError,
)
def download_raster(
    file_name,
    temp_dir,
    variable,
    bearer,
    quality_flag_rm,
    quiet,
    tile_i,
    n_tile,
    progress=None,
):
    """Download NASA BLA"""

    year = file_name[9:13]
    day = file_name[13:16]
    product_id = file_name[0:7]

    # f = os.path.join(temp_dir, product_id, year, day, file_name)
    f = os.path.join(temp_dir, file_name)

    url = f"https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/5000/{product_id}/{year}/{day}/{file_name}"
    headers = {"Authorization": f"Bearer {bearer}"}
    download_path = os.path.join(temp_dir, file_name)

    with httpx.stream(
        "GET",
        url,
        headers=headers,
    ) as response:
        total = int(response.headers["Content-Length"])
        with open(download_path, "wb") as download_file, tqdm(
            total=total,
            unit="B",
            unit_scale=True,
            leave=None,
        ) as pbar:
            pbar.set_description(f"Processing {file_name}...")
            for chunk in response.iter_raw():
                download_file.write(chunk)
                pbar.update(len(chunk))

    # Convert to raster
    file_name_tif = re.sub(".h5", ".tif", file_name)

    file_to_raster(
        f,
        variable,
        os.path.join(temp_dir, "tif_files_tmp", file_name_tif),
        quality_flag_rm,
    )


def define_variable(variable, product_id):
    if variable is None:
        if product_id == "VNP46A1":
            variable = "DNB_At_Sensor_Radiance_500m"
        if product_id == "VNP46A2":
            variable = "Gap_Filled_DNB_BRDF-Corrected_NTL"
        if product_id in ["VNP46A3", "VNP46A4"]:
            variable = "NearNadir_Composite_Snow_Free"
    return variable


def define_date_name(date_i, product_id):
    if product_id in ["VNP46A1", "VNP46A2"]:
        date_name_i = "t" + re.sub("-", "_", date_i)

    if product_id == "VNP46A3":
        date_name_i = "t" + re.sub("-", "_", date_i)[:7]

    if product_id == "VNP46A4":
        date_name_i = "t" + re.sub("-", "_", date_i)[:4]

    return date_name_i


def bm_extract_i(
    roi_sf,
    product_id,
    date_i,
    bearer,
    variable,
    quality_flag_rm,
    aggregation_fun,
    check_all_tiles_exist,
    quiet,
    temp_dir,
):
    try:
        #### Extract data
        raster_path_i = bm_raster_i(
            roi_sf,
            product_id,
            date_i,
            bearer,
            variable,
            quality_flag_rm,
            check_all_tiles_exist,
            quiet,
            temp_dir,
        )

        with rasterio.open(raster_path_i) as src:
            raster_data = src.read(1)

        # nodata=src.nodata
        ntl_data = zonal_stats(
            roi_sf,
            raster_data,
            affine=src.transform,
            nodata=np.nan,
            masked=False,
            stats=aggregation_fun,
        )

        ntl_data_df = pd.DataFrame(ntl_data)
        ntl_data_df = ntl_data_df.add_prefix("ntl_")

        roi_df = pd.DataFrame(roi_sf.drop("geometry", axis=1))
        poly_ntl_df = pd.concat([roi_df, ntl_data_df], axis=1)
        poly_ntl_df["date"] = date_i

    except:
        logger.info(
            "Skipping " + str(date_i) + " due to error. Data may not be available.\n"
        )
        poly_ntl_df = pd.DataFrame()

    return poly_ntl_df


def bm_raster_i(
    roi_sf: geopandas.GeoDataFrame,
    product_id: str,
    date: str,
    bearer,
    variable,
    quality_flag_rm,
    check_all_tiles_exist,
    quiet,
    temp_dir,
):
    #### Prep files to download
    if not isinstance(date, datetime.date):
        date = datetime.datetime.strptime(date, "%Y-%m-%d").date()

    bm_files_df = retrieve_manifest(product_id, date)
    bm_tiles_sf = gpd.read_file(
        files("blackmarble.data").joinpath("blackmarbletiles.geojson")
    )

    # Intersecting tiles
    bm_tiles_sf = bm_tiles_sf[~bm_tiles_sf["TileID"].str.contains("h00")]
    bm_tiles_sf = bm_tiles_sf[~bm_tiles_sf["TileID"].str.contains("v00")]

    grid_use_sf = gpd.overlay(bm_tiles_sf, roi_sf.dissolve(), how="intersection")

    # Make raster
    tile_ids_rx = "|".join(grid_use_sf["TileID"])
    bm_files_df = bm_files_df[bm_files_df["name"].str.contains(tile_ids_rx)]
    bm_files_df = bm_files_df.reset_index()

    #### Create directory for tif files
    shutil.rmtree(os.path.join(temp_dir, "tif_files_tmp"), ignore_errors=True)
    os.makedirs(os.path.join(temp_dir, "tif_files_tmp"))

    #### Download files and convert to rasters
    if (bm_files_df.shape[0] < grid_use_sf.shape[0]) and check_all_tiles_exist:
        logger.info(
            "Not all satellite imagery tiles for this location exist, so skipping. To ignore this error and process anyway, set check_all_tiles_exist = False"
        )
        raise ValueError(
            "Not all satellite imagery tiles for this location exist, so skipping. To ignore this error and process anyway, set check_all_tiles_exist = False"
        )

    else:
        tile_i = 1

        pbar = tqdm(bm_files_df["name"], leave=None)
        for file_name in pbar:
            n_tile = bm_files_df.shape[0]
            # Saves files in {temp_dir}/tif_files_tmp, which above is cleared and created
            download_raster(
                file_name,
                temp_dir,
                variable,
                bearer,
                quality_flag_rm,
                quiet,
                tile_i,
                n_tile,
            )
            tile_i = tile_i + 1

        #### Mosaic together
        # List of raster files to be mosaiced
        filepaths = glob.glob(os.path.join(temp_dir, "tif_files_tmp", "*.tif"))

        # Open the raster files
        src_files_to_mosaic = []
        for fp in filepaths:
            src = rasterio.open(fp)
            src_files_to_mosaic.append(src)

        # Merge the rasters
        mosaic, out_trans = merge(src_files_to_mosaic)

        # Delete folder of individual files
        # shutil.rmtree(os.path.join(temp_dir, 'tif_files_tmp'), ignore_errors=True)

        #### Create directory for mosaiced tif files
        # shutil.rmtree(os.path.join(temp_dir, 'tif_files_mosaic_tmp'), ignore_errors=True)
        os.makedirs(os.path.join(temp_dir, "tif_files_mosaic_tmp"), exist_ok=True)

        out_name = file_name[0:16] + ".tif"
        out_fp = os.path.join(temp_dir, "tif_files_mosaic_tmp", out_name)

        # Output as raster
        out_meta = src.meta.copy()

        out_meta.update(
            {
                "driver": "GTiff",
                "height": mosaic.shape[1],
                "width": mosaic.shape[2],
                "transform": out_trans,
                "crs": "+proj=longlat +datum=WGS84 +no_defs",
            }
        )

        with rasterio.open(out_fp, "w", **out_meta) as dest:
            dest.write(mosaic)

        #### Mask
        if True:
            dataset = rasterio.open(out_fp)
            mask_geometry = roi_sf.dissolve().geometry.values[
                0
            ]  # roi_sf.geometry.values[0]

            masked_image, mask_transform = mask(
                dataset,
                [mask_geometry],
                crop=True,
                nodata=np.nan,
                all_touched=True,
                pad=True,
            )

            # Copy the metadata from the original raster dataset
            masked_meta = dataset.meta.copy()
            masked_meta.update(
                {
                    "driver": "GTiff",
                    "height": masked_image.shape[1],
                    "width": masked_image.shape[2],
                    "transform": mask_transform,
                }
            )

            # Export the masked raster to a new file
            with rasterio.open(out_fp, "w", **masked_meta) as masked_dataset:
                masked_dataset.write(masked_image)

    return out_fp



import os
import shutil
import tempfile
import time

import numpy as np
import rasterio
from rasterio.merge import merge

from tqdm.auto import tqdm

from . import logger
from .utils import bm_raster_i, define_date_name, define_variable


def bm_raster(
    roi_sf,
    product_id,
    date,
    bearer,
    variable=None,
    quality_flag_rm=[255],
    check_all_tiles_exist=True,
    output_location_type="tempfile",
    file_dir=None,
    file_prefix="",
    file_skip_if_exists=True,
    quiet=False,
):
    """Make a raster of nighttime lights from [NASA Black Marble data](https://blackmarble.gsfc.nasa.gov/).

    Parameters
    ----------

    roi_st : geopandas.DataFrame
        Region of interest; geopandas dataframe (polygon). Must be in the [WGS 84 (epsg:4326)](https://epsg.io/4326) coordinate reference system.

    product_id: string. One of the following:
        * `"VNP46A1"`: Daily (raw)
        * `"VNP46A2"`: Daily (corrected)
        * `"VNP46A3"`: Monthly
        * `"VNP46A4"`: Annual

    date: Date of raster data. Entering one date will produce a raster. Entering multiple dates will produce a raster stack.
        * For `product_id`s `"VNP46A1"` and `"VNP46A2"`, a date (eg, `"2021-10-03"`).
        * For `product_id` `"VNP46A3"`, a date or year-month (e.g., `"2021-10-01"`, where the day will be ignored, or `"2021-10"`).
        * For `product_id` `"VNP46A4"`, year or date  (e.g., `"2021-10-01"`, where the month and day will be ignored, or `2021`).

    bearer: NASA bearer token. For instructions on how to create a token, see [here](https://github.com/ramarty/blackmarbler#bearer-token-).

    variable: Variable to used to create raster (default: `NULL`). If `NULL`, uses the following default variables:
        * For `product_id` `:VNP46A1"`, uses `DNB_At_Sensor_Radiance_500m`.
        * For `product_id` `"VNP46A2"`, uses `Gap_Filled_DNB_BRDF-Corrected_NTL`.
        * For `product_id`s `"VNP46A3"` and `"VNP46A4"`, uses `NearNadir_Composite_Snow_Free`.
    For information on other variable choices, see [here](https://ladsweb.modaps.eosdis.nasa.gov/api/v2/content/archives/Document%20Archive/Science%20Data%20Product%20Documentation/VIIRS_Black_Marble_UG_v1.2_April_2021.pdf); for `VNP46A1`, see Table 3; for `VNP46A2` see Table 6; for `VNP46A3` and `VNP46A4`, see Table 9.

    quality_flag_rm: long
        Quality flag values to use to set values to `NA`. Each pixel has a quality flag value, where low quality values can be removed. Values are set to `NA` for each value in ther `quality_flag_rm` vector. (Default: `c(255)`).

        For `VNP46A1` and `VNP46A2` (daily data):
        - `0`: High-quality, Persistent nighttime lights
        - `1`: High-quality, Ephemeral nighttime Lights
        - `2`: Poor-quality, Outlier, potential cloud contamination, or other issues
        - `255`: No retrieval, Fill value (masked out on ingestion)

    For `VNP46A3` and `VNP46A4` (monthly and annual data):
    - `0`: Good-quality, The number of observations used for the composite is larger than 3
    - `1`: Poor-quality, The number of observations used for the composite is less than or equal to 3
    - `2`: Gap filled NTL based on historical data
    - `255`: Fill value

    check_all_tiles_exist: Check whether all Black Marble nighttime light tiles exist for the region of interest. Sometimes not all tiles are available, so the full region of interest may not be covered. If `TRUE`, skips cases where not all tiles are available. (Default: `TRUE`).

    output_location_type: Where to produce output; either `r_memory` or `file`. If `r_memory`, functions returns a raster in R. If `file`, function exports a `.tif` file and returns `NULL`.

    For `output_location_type = file`:

        file_dir: The directory where data should be exported (default: `NULL`, so the working directory will be used)

        file_prefix: Prefix to add to the file to be saved. The file will be saved as the following: `[file_prefix][product_id]_t[date].tif`

        file_skip_if_exists: Whether the function should first check wither the file already exists, and to skip downloading or extracting data if the data for that date if the file already exists (default: `TRUE`).

    quiet: Suppress output that show downloading progress and other messages. (Default: `FALSE`).

    Returns
    -------
    None (if output_location_type = "file")
        A geotif file is saved to the "file_dir" directory. Nothing is returned from the function.

    Raster (if output_location_type = "tempfile")

    """

    #### Make directory to put temporary files into
    temp_main_dir = tempfile.gettempdir()

    current_time_millis = int(round(time.time() * 1000))
    temp_dir = os.path.join(temp_main_dir, "bm_raster_temp_" + str(current_time_millis))

    os.makedirs(temp_dir)

    #### Define NTL Variable
    variable = define_variable(variable, product_id)

    #### Ensure quality_flag_rm is a list
    if type(quality_flag_rm) is not list:
        quality_flag_rm = [quality_flag_rm]

    #### Ensure date is a list of strings
    if type(date) is not list:
        date = [date]

    date = [str(item) for item in date]

    if file_dir is None:
        file_dir = os.getcwd()

    # File --------------------------------------------------------------------------
    if output_location_type == "file":
        # crate progress bar
        pbar = tqdm(date, leave=None)

        for date_i in pbar:
            pbar.set_description("Downloading raster(s)...")
            date_name_i = define_date_name(date_i, product_id)

            try:
                out_name = file_prefix + product_id + "_" + date_name_i + ".tif"
                out_path = os.path.join(file_dir, out_name)

                # Only make .tif if raster doesn't already exist
                if (not file_skip_if_exists) | (not os.path.exists(out_path)):
                    raster_path_i = bm_raster_i(
                        roi_sf,
                        product_id,
                        date_i,
                        bearer,
                        variable,
                        quality_flag_rm,
                        check_all_tiles_exist,
                        quiet,
                        temp_dir,
                    )
                    shutil.move(raster_path_i, out_path)  # Move from tmp to main folder

                    if quiet == False:
                        logger.info("File created: " + out_path)

                else:
                    if quiet == False:
                        logger.info('"' + out_path + '" already exists; skipping.\n')

                r_out = None
            except:
                r_out = None

    # Tempfile --------------------------------------------------------------------------
    if output_location_type == "tempfile":
        try:
            #### Create raster for each date
            raster_path_list = []
            for date_i in date:
                date_name_i = define_date_name(date_i, product_id)

                raster_path_i = bm_raster_i(
                    roi_sf,
                    product_id,
                    date_i,
                    bearer,
                    variable,
                    quality_flag_rm,
                    check_all_tiles_exist,
                    quiet,
                    temp_dir,
                )
                raster_path_list.append(raster_path_i)

            #### Stack Rasters
            # Read the first file to get the dimensions and metadata
            with rasterio.open(raster_path_list[0]) as src:
                width = src.width
                height = src.height
                count = len(raster_path_list)
                crs = src.crs
                transform = src.transform
                dtype = src.dtypes[0]

                # Create an empty numpy array to store the raster data
                data = np.zeros((count, height, width), dtype=dtype)

                # Read data from each file and store it in the numpy array
                for i, file_path in enumerate(raster_path_list):
                    with rasterio.open(file_path) as src:
                        data[i] = src.read(1)

            # Create a new raster file and write the data
            timestamp = str(int(time.time()))
            tmp_raster_file_name = (
                product_id + "_" + date[0].replace("-", "_") + "_" + timestamp + ".tif"
            )

            with rasterio.open(
                os.path.join(temp_dir, tmp_raster_file_name),
                "w",
                driver="GTiff",
                width=width,
                height=height,
                count=count,
                dtype=dtype,
                crs=crs,
                transform=transform,
            ) as dst:
                for i in range(count):
                    dst.write(data[i], i + 1)

            # Move from black marble temp folder to main temp folder
            shutil.move(os.path.join(temp_dir, tmp_raster_file_name), temp_main_dir)

            r_out = rasterio.open(os.path.join(temp_main_dir, tmp_raster_file_name))

        except Exception as e:
            # Delete temp files used to make raster
            shutil.rmtree(temp_dir, ignore_errors=True)

            if quiet == False:
                logger.info(
                    "Skipping "
                    + str(date_i)
                    + " due to error. Data may not be available.\n"
                )
            r_out = None

    # Delete temp files used to make raster
    shutil.rmtree(temp_dir, ignore_errors=True)

    return r_out



from pathlib import Path
import logging

# from .base import BlackMarbleDownloader  # noqa: F401

__version__ = (Path(__file__).parent / "VERSION").read_text().strip()

logging.basicConfig(
    format="[%(asctime)s - %(name)s:%(lineno)d - %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

logger: logging.Logger = logging.getLogger(__name__)

logger.setLevel(logging.DEBUG)



import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

import httpx
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

from .utils import file_to_raster

# example constant variable
NAME = "blackmarble"


class DownloadProgress(Progress):
    def get_renderables(self):
        for task in self.tasks:
            if task.fields.get("progress_type") == "summary":
                self.columns = (
                    TextColumn(
                        "[blue]{task.description}[/blue][magenta]{task.completed} of {task.total}...",
                        justify="right",
                    ),
                    BarColumn(bar_width=None),
                    TimeRemainingColumn(),
                    TimeElapsedColumn(),
                )
            if task.fields.get("progress_type") == "download":
                self.columns = (
                    TextColumn("[blue]{task.description}", justify="left"),
                    DownloadColumn(),
                    BarColumn(bar_width=None),
                    "[progress.percentage]{task.percentage:>3.0f}%",
                )
            yield self.make_tasks_table([task])


@dataclass
class BlackMarbleDownloader:
    """NASA Black Marble Downloader

    Attributes
    ----------
    bearer: str
        NASA bearer
    """

    bearer: Optional[str]
    cache_dir: Optional[Union[str, Path]] = None

    def download_raster(
        self,
        file_name,
        temp_dir,
        variable,
        quality_flag_rm,
        progress=DownloadProgress(),
    ):
        # Path
        year = file_name[9:13]
        day = file_name[13:16]
        product_id = file_name[0:7]

        f = os.path.join(temp_dir, file_name)

        url = f"https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/5000/{product_id}/{year}/{day}/{file_name}"
        headers = {"Authorization": f"Bearer {self.bearer}"}
        download_path = os.path.join(temp_dir, file_name)

        with httpx.stream("GET", url, headers=headers) as response:
            total = int(response.headers["Content-Length"])
            with progress:
                with open(download_path, "wb") as download_file:
                    task = progress.add_task(
                        "[red]Downloading...[/red]",
                        total=total,
                        progress_type="download",
                    )
                    for chunk in response.iter_raw():
                        download_file.write(chunk)
                        progress.update(
                            task,
                            advance=len(chunk),
                            description=f"[yellow]Downloading {file_name}...[/yellow]",
                        )

        file_name_tif = re.sub(".h5", ".tif", file_name)

        file_to_raster(
            f,
            variable,
            os.path.join(temp_dir, "tif_files_tmp", file_name_tif),
            quality_flag_rm,
        )

        progress.update(
            task,
            description=f"[green]Downloaded {file_name}[/green]",
        )
        return None
